import os
import unittest

from lxml import etree
from xmldiff.formatting import DiffFormatter, WS_NONE
from xmldiff.main import diff_trees, diff_texts, patch_text, patch_file
from xmldiff.patch import Patcher, DiffParser
from xmldiff.actions import (UpdateTextIn, InsertNode, MoveNode,
                             DeleteNode, UpdateAttrib, InsertAttrib,
                             RenameAttrib, DeleteAttrib, UpdateTextAfter,
                             RenameNode, InsertComment)

from .testing import compare_elements


class PatcherTests(unittest.TestCase):

    patcher = Patcher()

    def _test(self, start, action, end):
        tree = etree.fromstring(start)
        self.patcher.handle_action(action, tree)
        self.assertEqual(etree.tounicode(tree), end)

    def test_delete_node(self):
        self._test('<root><deleteme/></root>',
                   DeleteNode('/root/deleteme'),
                   '<root/>')

    def test_insert_node(self):
        self._test('<root><anode/></root>',
                   InsertNode('/root/anode', 'newnode', 0),
                   '<root><anode><newnode/></anode></root>')

    def test_rename_node(self):
        self._test('<root><oldname/></root>',
                   RenameNode('/root/oldname', 'newname'),
                   '<root><newname/></root>')

    def test_move_node(self):
        self._test('<root><anode><moveme/></anode></root>',
                   MoveNode('/root/anode/moveme', '/root', 1),
                   '<root><anode/><moveme/></root>')

    def test_update_text_in(self):
        self._test('<root><anode/></root>',
                   UpdateTextIn('/root/anode', 'New text'),
                   '<root><anode>New text</anode></root>')

    def test_update_text_after(self):
        self._test('<root><anode/></root>',
                   UpdateTextAfter('/root/anode', 'New text'),
                   '<root><anode/>New text</root>')

    def test_update_attrib(self):
        self._test('<root><anode attrib="oldvalue" /></root>',
                   UpdateAttrib('/root/anode', 'attrib', 'newvalue'),
                   '<root><anode attrib="newvalue"/></root>')

    def test_delete_attrib(self):
        self._test('<root><anode attrib="oldvalue" /></root>',
                   DeleteAttrib('/root/anode', 'attrib'),
                   '<root><anode/></root>')

    def test_insert_attrib(self):
        self._test('<root><anode/></root>',
                   InsertAttrib('/root/anode', 'attrib', 'value'),
                   '<root><anode attrib="value"/></root>')

    def test_rename_attrib(self):
        self._test('<root><anode oldname="value"/></root>',
                   RenameAttrib('/root/anode', 'oldname', 'newname'),
                   '<root><anode newname="value"/></root>')

    def test_insert_comment(self):
        self._test('<root><anode/></root>',
                   InsertComment('/root', 1, "This is a new comment"),
                   '<root><anode/><!--This is a new comment--></root>')


class DiffPatch(unittest.TestCase):

    def test_diff_patch(self):
        here = os.path.split(__file__)[0]
        lfile = os.path.join(here, 'test_data', 'all_actions.left.xml')
        rfile = os.path.join(here, 'test_data', 'all_actions.right.xml')

        left = etree.parse(lfile)
        right = etree.parse(rfile)
        diff = diff_trees(left, right)
        result = Patcher().patch(diff, left)

        # This example has top level comments, and lxml doesn't deal well
        # with that, so the trees are not EXACTLY the same, the trailing
        # top level comment differs, but that's OK.
        compare_elements(result.getroot(), right.getroot())


TEST_DIFF = '''[delete, node]
[insert, target, tag, 0]
[rename, node, tag]
[move, node, target, 0]
[update-text, node, "text"]
[update-text-after, node, "text"]
[update-attribute, node, name, "value"]
[delete-attribute, node, name]
[insert-attribute, node, name, "value"]
[rename-attribute, node, oldname, newname]
[insert-comment, target, 0, "text"]
'''


class ParserTests(unittest.TestCase):

    def test_make_action(self):
        parser = DiffParser()

        self.assertEqual(
            parser.make_action('[delete, node]'),
            DeleteNode('node')
        )

        self.assertEqual(
            parser.make_action('[insert, target, tag, 0]'),
            InsertNode('target', 'tag', 0)
        )

        self.assertEqual(
            parser.make_action('[rename, node, tag]'),
            RenameNode('node', 'tag')
        )

        self.assertEqual(
            parser.make_action('[move, node, target, 0]'),
            MoveNode('node', 'target', 0)
        )

        self.assertEqual(
            parser.make_action('[update-text, node, "text"]'),
            UpdateTextIn('node', 'text')
        )

        self.assertEqual(
            parser.make_action('[update-text-after, node, "text"]'),
            UpdateTextAfter('node', 'text')
        )

        self.assertEqual(
            parser.make_action('[update-attribute, node, name, "value"]'),
            UpdateAttrib('node', 'name', 'value')
        )

        self.assertEqual(
            parser.make_action('[delete-attribute, node, name]'),
            DeleteAttrib('node', 'name')
        )

        self.assertEqual(
            parser.make_action('[insert-attribute, node, name, "value"]'),
            InsertAttrib('node', 'name', 'value')
        )

        self.assertEqual(
            parser.make_action('[rename-attribute, node, oldname, newname]'),
            RenameAttrib('node', 'oldname', 'newname')
        )

        self.assertEqual(
            parser.make_action('[insert-comment, target, 0, "text"]'),
            InsertComment('target', 0, 'text')
        )

    def test_parse(self):
        parser = DiffParser()
        actions = list(parser.parse(TEST_DIFF))
        self.assertEqual(len(actions), len(TEST_DIFF.splitlines()))

    def test_parse_broken(self):
        # Testing incorrect patch files
        parser = DiffParser()

        # Empty file, nothing happens
        actions = list(parser.parse(''))
        self.assertEqual(actions, [])

        # Not a diff raises error
        with self.assertRaises(ValueError):
            actions = list(parser.parse('Not a diff'))

        # It should handle lines that have been broken, say in an email
        actions = list(parser.parse('[insert-comment, target,\n 0, "text"]'))
        self.assertEqual(actions, [InsertComment('target', 0, 'text')])

        # It should not handle broken files
        with self.assertRaises(ValueError):
            actions = list(parser.parse('[insert-comment, target,\n'))

    def test_diff_patch(self):
        here = os.path.split(__file__)[0]
        lfile = os.path.join(here, 'test_data', 'all_actions.left.xml')
        rfile = os.path.join(here, 'test_data', 'all_actions.right.xml')
        with open(lfile) as f:
            left = f.read()
        with open(rfile) as f:
            right = f.read()

        diff = diff_texts(left, right,
                          formatter=DiffFormatter(normalize=WS_NONE))
        result = patch_text(diff, left)
        compare_elements(etree.fromstring(result), etree.fromstring(right))

    def test_patch_stream(self):
        here = os.path.join(os.path.split(__file__)[0], 'test_data')
        xmlfile = os.path.join(here, 'insert-node.left.html')
        patchfile = os.path.join(here, 'insert-node.diff')
        result = patch_file(patchfile, xmlfile)

        expectedfile = os.path.join(here, 'insert-node.right.html')
        with open(expectedfile, 'rt') as f:
            expected = f.read()
        # lxml.etree.parse() will strip ending whitespace
        self.assertEqual(result, expected.rstrip())



from collections import namedtuple

# The edit script actions used in xmldiff
DeleteNode = namedtuple('DeleteNode', 'node')
InsertNode = namedtuple('InsertNode', 'target tag position')
RenameNode = namedtuple('RenameNode', 'node tag')
MoveNode = namedtuple('MoveNode', 'node target position')

UpdateTextIn = namedtuple('UpdateTextIn', 'node text')
UpdateTextAfter = namedtuple('UpdateTextAfter', 'node text')

UpdateAttrib = namedtuple('UpdateAttrib', 'node name value')
DeleteAttrib = namedtuple('DeleteAttrib', 'node name')
InsertAttrib = namedtuple('InsertAttrib', 'node name value')
RenameAttrib = namedtuple('RenameAttrib', 'node oldname newname')

InsertComment = namedtuple('InsertComment', 'target position text')



"""All major API points and command-line tools"""
import pkg_resources
import six

from argparse import ArgumentParser, FileType
from lxml import etree
from xmldiff import diff, formatting, patch

__version__ = pkg_resources.require("xmldiff")[0].version

FORMATTERS = {
    'diff': formatting.DiffFormatter,
    'xml': formatting.XMLFormatter,
    'old': formatting.XmlDiffFormatter,
}


def diff_trees(left, right, diff_options=None, formatter=None):
    """Takes two lxml root elements or element trees"""
    if formatter is not None:
        formatter.prepare(left, right)
    if diff_options is None:
        diff_options = {}
    differ = diff.Differ(**diff_options)
    diffs = differ.diff(left, right)

    if formatter is None:
        return list(diffs)

    return formatter.format(diffs, left)


def _diff(parse_method, left, right, diff_options=None, formatter=None):
    normalize = bool(getattr(formatter, 'normalize', 1) & formatting.WS_TAGS)
    parser = etree.XMLParser(remove_blank_text=normalize)
    left_tree = parse_method(left, parser)
    right_tree = parse_method(right, parser)
    return diff_trees(left_tree, right_tree, diff_options=diff_options,
                      formatter=formatter)


def diff_texts(left, right, diff_options=None, formatter=None):
    """Takes two Unicode strings containing XML"""
    return _diff(etree.fromstring, left, right,
                 diff_options=diff_options, formatter=formatter)


def diff_files(left, right, diff_options=None, formatter=None):
    """Takes two filenames or streams, and diffs the XML in those files"""
    return _diff(etree.parse, left, right,
                 diff_options=diff_options, formatter=formatter)


def make_diff_parser():
    parser = ArgumentParser(description='Create a diff for two XML files.',
                            add_help=False)
    parser.add_argument('file1', type=FileType('r'),
                        help='The first input file.')
    parser.add_argument('file2', type=FileType('r'),
                        help='The second input file.')
    parser.add_argument('-h', '--help', action='help',
                        help='Show this help message and exit.')
    parser.add_argument('-v', '--version', action='version',
                        help='Display version and exit.',
                        version='xmldiff %s' % __version__)
    parser.add_argument('-f', '--formatter', default='diff',
                        choices=list(FORMATTERS.keys()),
                        help='Formatter selection.')
    parser.add_argument('-w', '--keep-whitespace', action='store_true',
                        help='Do not strip ignorable whitespace.')
    parser.add_argument('-p', '--pretty-print', action='store_true',
                        help='Try to make XML output more readable.')
    parser.add_argument('-F', type=float,
                        help='A value between 0 and 1 that determines how '
                        'similar nodes must be to match.')
    parser.add_argument('--unique-attributes', type=str, nargs='?',
                        default='{http://www.w3.org/XML/1998/namespace}id',
                        help='A comma separated list of attributes '
                             'that uniquely identify a node. Can be empty. '
                             'Unique attributes for certain elements can '
                             'be specified in the format {NS}element@attr.')
    parser.add_argument('--ratio-mode', default='fast',
                        choices={'accurate', 'fast', 'faster'},
                        help='Choose the node comparison optimization.')
    parser.add_argument('--fast-match', action='store_true',
                        help='A faster, less optimal match run.')
    return parser


def _parse_uniqueattrs(uniqueattrs):
    if uniqueattrs is None:
        return []
    return [
        attr if '@' not in attr else attr.split('@', 1)
        for attr in uniqueattrs.split(',')
    ]


def diff_command(args=None):
    parser = make_diff_parser()
    args = parser.parse_args(args=args)

    if args.keep_whitespace:
        normalize = formatting.WS_NONE
    else:
        normalize = formatting.WS_BOTH

    formatter = FORMATTERS[args.formatter](normalize=normalize,
                                           pretty_print=args.pretty_print)

    diff_options = {'ratio_mode': args.ratio_mode,
                    'F': args.F,
                    'fast_match': args.fast_match,
                    'uniqueattrs': _parse_uniqueattrs(args.unique_attributes),
                    }
    result = diff_files(args.file1, args.file2, diff_options=diff_options,
                        formatter=formatter)
    print(result)


def patch_tree(actions, tree):
    """Takes an lxml root element or element tree, and a list of actions"""
    patcher = patch.Patcher()
    return patcher.patch(actions, tree)


def patch_text(actions, tree):
    """Takes a string with XML and a string with actions"""
    tree = etree.fromstring(tree)
    actions = patch.DiffParser().parse(actions)
    tree = patch_tree(actions, tree)
    return etree.tounicode(tree)


def patch_file(actions, tree):
    """Takes two filenames or streams, one with XML the other a diff"""
    tree = etree.parse(tree)

    if isinstance(actions, six.string_types):
        # It's a string, so it's a filename
        with open(actions) as f:
            actions = f.read()
    else:
        # We assume it's a stream
        actions = actions.read()

    actions = patch.DiffParser().parse(actions)
    tree = patch_tree(actions, tree)
    return etree.tounicode(tree)


def make_patch_parser():
    parser = ArgumentParser(description='Patch an XML file with an xmldiff',
                            add_help=False)
    parser.add_argument('patchfile', type=FileType('r'),
                        help='An xmldiff diff file.')
    parser.add_argument('xmlfile', type=FileType('r'),
                        help='An unpatched XML file.')
    parser.add_argument('-h', '--help', action='help',
                        help='Show this help message and exit.')
    parser.add_argument('-v', '--version', action='version',
                        help='Display version and exit.',
                        version='xmldiff %s' % __version__)
    return parser


def patch_command(args=None):
    parser = make_patch_parser()
    args = parser.parse_args(args=args)

    result = patch_file(args.patchfile, args.xmlfile)
    print(result)
